# coding: utf-8
from typing import Any, Dict, Optional

from delta_trace_db.query.cause.permission import Permission
from delta_trace_db.query.query import Query
from delta_trace_db.query.transaction_query import TransactionQuery


class UtilQuery:
    """
    (en) Utilities for query processing.
    (ja) クエリ処理用のユーティリティです。
    """

    @staticmethod
    def convert_from_json(src: Dict[str, Any]) -> Any:
        """
        (en) Restores a Query or TransactionQuery class from a JSON dict.
        (ja) JSONのdictから、QueryまたはTransactionQueryクラスを復元します。

        Args:
            src (Dict[str, Any]): The dict of Query or TransactionQuery class.

        Returns:
            Query | TransactionQuery

        Raises:
            ValueError: If you pass an incorrect class.
        """
        try:
            if src.get("className") == "Query":
                return Query.from_dict(src)
            elif src.get("className") == "TransactionQuery":
                return TransactionQuery.from_dict(src)
            else:
                raise ValueError(f"Unsupported query class: {src.get('className')}")
        except Exception:
            raise ValueError("Unsupported object")

    @staticmethod
    def check_permissions(q: Query, collection_permissions: Optional[Dict[str, Permission]]) -> bool:
        if collection_permissions is None:
            return True
        if q.target not in collection_permissions:
            return False

        # allowsのチェック
        p = collection_permissions[q.target]
        if q.type in p.allows:
            return True

        return False



# coding: utf-8
from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.cause.permission import Permission
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.query_builder import QueryBuilder
from delta_trace_db.query.transaction_query import TransactionQuery


def test_permission():
    db = DeltaTraceDatabase()

    q = QueryBuilder.clear(
        target="user",
        must_affect_at_least_one=False
    ).build()

    tq = TransactionQuery(queries=[q])

    r1 = db.execute_query(q, collection_permissions=None)
    r2 = db.execute_query(q, collection_permissions={})
    r3 = db.execute_query(q, collection_permissions={"user": Permission([])})
    r4 = db.execute_query(q, collection_permissions={"user": Permission([EnumQueryType.add])})
    r5 = db.execute_query(q, collection_permissions={"user": Permission([EnumQueryType.clear])})
    r6 = db.execute_query(q, collection_permissions={"item": Permission([EnumQueryType.clear])})

    tr1 = db.execute_transaction_query(tq, collection_permissions=None)
    tr2 = db.execute_transaction_query(tq, collection_permissions={})
    tr3 = db.execute_transaction_query(tq, collection_permissions={"user": Permission([])})
    tr4 = db.execute_transaction_query(tq, collection_permissions={"user": Permission([EnumQueryType.add])})
    tr5 = db.execute_transaction_query(tq, collection_permissions={"user": Permission([EnumQueryType.clear])})
    tr6 = db.execute_transaction_query(tq, collection_permissions={"item": Permission([EnumQueryType.clear])})

    assert r1.is_success is True
    assert r2.is_success is False
    assert r3.is_success is False
    assert r4.is_success is False
    assert r5.is_success is True
    assert r6.is_success is False

    assert tr1.is_success is True
    assert tr2.is_success is False
    assert tr3.is_success is False
    assert tr4.is_success is False
    assert tr5.is_success is True
    assert tr6.is_success is False



# coding: utf-8
from typing import Any, Dict

class UtilField:
    """
    (en) Utility for field access used in internal DB processing.
    (ja) DBの内部処理で利用する、フィールドアクセスに関するユーティリティです。
    """

    @staticmethod
    def get_nested_field_value(map_: Dict[str, Any], path: str) -> object | None:
        """
        (en) Access nested fields of a dictionary.
        (ja) 辞書の、ネストされたフィールドにアクセスするための関数です。

        Args:
            map_ (Dict[str, Any]): 探索したいマップ。
            path (str): "."区切りの探索用パス。例: "user.name"

        Returns:
            object | None: 見つかった値、または存在しなければ None
        """
        keys = path.split('.')
        current: object | None = map_
        for key in keys:
            if isinstance(current, dict) and key in current:
                current = current[key]
            else:
                return None
        return current



# coding: utf-8
from enum import Enum

class EnumNodeType(Enum):
    # 論理演算ノード (構造的に条件を組み立てる)
    and_ = "and_"
    or_ = "or_"
    not_ = "not_"

    # 比較/条件ノード (フィールド値に対して条件を設定)
    equals_ = "equals_"
    notEquals_ = "notEquals_"
    greaterThan_ = "greaterThan_"
    lessThan_ = "lessThan_"
    greaterThanOrEqual_ = "greaterThanOrEqual_"
    lessThanOrEqual_ = "lessThanOrEqual_"
    regex_ = "regex_"
    contains_ = "contains_"
    in_ = "in_"
    notIn_ = "notIn_"
    startsWith_ = "startsWith_"
    endsWith_ = "endsWith_"



# coding: utf-8
from typing import Any

class UtilCopy:
    _max_depth = 100  # 安全な再帰深度上限

    @staticmethod
    def jsonable_deep_copy(value: Any, depth: int = 0) -> Any:
        """
        JSON互換型のみを再帰的にディープコピーします。
        非対応の型や再帰深度超過は ValueError を投げます。

        Args:
            value: コピー対象
            depth: 内部用の再帰深度カウンタ（外部から設定不要）

        Returns:
            コピーされたオブジェクト
        """
        if depth > UtilCopy._max_depth:
            raise ValueError(f'Exceeded max allowed nesting depth ({UtilCopy._max_depth})')

        if isinstance(value, dict):
            return {k: UtilCopy.jsonable_deep_copy(v, depth=depth+1) for k, v in value.items()}
        elif isinstance(value, list):
            return [UtilCopy.jsonable_deep_copy(v, depth=depth+1) for v in value]
        elif isinstance(value, (str, int, float, bool)) or value is None:
            return value
        else:
            raise ValueError(f'Unsupported type for JSON deep copy: {type(value)}')



# coding: utf-8
import json
from datetime import datetime, timedelta

from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.cause.actor import Actor
from delta_trace_db.query.cause.cause import Cause
from delta_trace_db.query.cause.enum_actor_type import EnumActorType
from delta_trace_db.query.cause.permission import Permission
from delta_trace_db.query.cause.temporal_trace.temporal_trace import TemporalTrace
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.nodes.comparison_node import FieldStartsWith, FieldGreaterThanOrEqual, FieldLessThanOrEqual, \
    FieldEquals
from delta_trace_db.query.nodes.logical_node import AndNode, OrNode
from delta_trace_db.query.query import Query
from delta_trace_db.query.query_builder import QueryBuilder
from delta_trace_db.query.query_result import QueryResult
from file_state_manager.cloneable_file import CloneableFile

from delta_trace_db.query.sort.multi_sort import MultiSort
from delta_trace_db.query.sort.single_sort import SingleSort


class User(CloneableFile):
    def __init__(self, id_, name, age, created_at, updated_at, nested_obj):
        super().__init__()
        self.id = id_
        self.name = name
        self.age = age
        self.created_at = created_at
        self.updated_at = updated_at
        self.nested_obj = nested_obj

    @classmethod
    def from_dict(cls, src) -> "User":
        return User(
            id_=src["id"],
            name=src["name"],
            age=src["age"],
            created_at=datetime.fromisoformat(src["createdAt"]),
            updated_at=datetime.fromisoformat(src["updatedAt"]),
            nested_obj=src["nestedObj"],
        )

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "age": self.age,
            "createdAt": self.created_at.isoformat(),
            "updatedAt": datetime.now().isoformat(),
            "nestedObj": dict(self.nested_obj),
        }

    def clone(self):
        return User.from_dict(self.to_dict())


class User2(CloneableFile):
    def __init__(self, id_, name, age, gender):
        super().__init__()
        self.id = id_
        self.name = name
        self.age = age
        self.gender = gender

    @classmethod
    def from_dict(cls, src) -> "User2":
        return User2(
            id_=src["id"],
            name=src["name"],
            age=src["age"],
            gender=src["gender"],
        )

    def to_dict(self):
        return {"id": self.id, "name": self.name, "age": self.age, "gender": self.gender}

    def clone(self):
        return User2.from_dict(self.to_dict())


class User3(CloneableFile):
    def __init__(self, serial_id, name, age, gender):
        super().__init__()
        self.serial_id = serial_id
        self.name = name
        self.age = age
        self.gender = gender

    @classmethod
    def from_dict(cls, src) -> "User3":
        return User3(
            serial_id=src["serialID"],
            name=src["name"],
            age=src["age"],
            gender=src["gender"],
        )

    def to_dict(self):
        return {
            "serialID": self.serial_id,
            "name": self.name,
            "age": self.age,
            "gender": self.gender,
        }

    def clone(self):
        return User3.from_dict(self.to_dict())


class Item1(CloneableFile):
    def __init__(self, name, serial_key=None):
        super().__init__()
        self.serial_key = serial_key
        self.name = name

    @classmethod
    def from_dict(cls, src) -> "Item1":
        return Item1(
            serial_key=src["serialKey"],
            name=src["name"],
        )

    def to_dict(self):
        return {
            "serialKey": self.serial_key,
            "name": self.name,
        }

    def clone(self):
        return Item1.from_dict(self.to_dict())


def test_db_basic_operation_part1():
    now = datetime.now()
    db = DeltaTraceDatabase()
    users = [
        User("1", "サンプル太郎", 25, now, now, {}),
        User("2", "サンプル次郎", 28, now + timedelta(days=1), now + timedelta(days=1), {}),
        User("3", "サンプル三郎", 31, now + timedelta(days=2), now + timedelta(days=2), {}),
        User("4", "サンプル花子", 17, now + timedelta(days=3), now + timedelta(days=3), {}),
    ]

    # add
    q1 = QueryBuilder.add(target="users", add_data=users).build()
    r1 = db.execute_query(Query.from_dict(q1.to_dict()))
    r1 = QueryResult.from_dict(r1.to_dict())
    assert r1.is_success
    assert r1.db_length == 4

    # clearAdd
    q1ex = QueryBuilder.clear_add(target="users", add_data=users).build()
    r1ex = db.execute_query(Query.from_dict(q1ex.to_dict()))
    assert r1ex.is_success
    assert r1ex.db_length == 4

    # search
    q2 = QueryBuilder.search(
        target="users",
        query_node=FieldStartsWith("name", "サンプル"),
        sort_obj=SingleSort(field="age", reversed_=True),
        limit=2,
    ).build()
    r2 = db.execute_query(Query.from_dict(q2.to_dict()))
    assert r2.db_length == 4
    assert r2.hit_count == 4
    result2 = r2.convert(User.from_dict)
    assert len(result2) == 2
    assert result2[0].name == "サンプル三郎"

    # query result serialize
    r2s = db.execute_query(q2)
    resume_result = QueryResult.from_dict(json.loads(json.dumps(r2s.to_dict())))
    assert resume_result.db_length == 4
    assert resume_result.hit_count == 4
    result2_resume = resume_result.convert(User.from_dict)
    assert len(result2_resume) == 2
    assert result2_resume[0].name == "サンプル三郎"

    # search by datetime
    q2dt = QueryBuilder.search(
        target="users",
        query_node=AndNode(
            [
                FieldGreaterThanOrEqual("createdAt", now + timedelta(days=1)),
                FieldLessThanOrEqual("createdAt", now + timedelta(days=2)),
            ]
        ),
        sort_obj=SingleSort(field="createdAt", reversed_=True),
    ).build()
    r2dt = db.execute_query(Query.from_dict(q2dt.to_dict()))
    assert r2dt.db_length == 4
    assert r2dt.hit_count == 2
    result2_dt = r2dt.convert(User.from_dict)
    assert len(result2_dt) == 2
    assert result2_dt[0].name == "サンプル三郎"
    assert result2_dt[1].name == "サンプル次郎"

    # paging
    q3 = QueryBuilder.search(
        target="users",
        query_node=FieldStartsWith("name", "サンプル"),
        sort_obj=SingleSort(field="age", reversed_=True),
        limit=2,
        start_after=r2.result[-1],
    ).build()
    r3 = db.execute_query(Query.from_dict(q3.to_dict()))
    assert r3.db_length == 4
    assert r3.hit_count == 4
    result3 = r3.convert(User.from_dict)
    assert len(result3) == 2
    assert result3[0].name == "サンプル太郎"
    assert result3[1].name == "サンプル花子"

    # pagingByOffset
    q3_offset = QueryBuilder.search(
        target="users",
        query_node=FieldStartsWith("name", "サンプル"),
        sort_obj=SingleSort(field="age", reversed_=True),
        limit=2,
        offset=2,
    ).build()
    r3_offset = db.execute_query(Query.from_dict(q3_offset.to_dict()))
    assert r3_offset.db_length == 4
    assert r3_offset.hit_count == 4
    result3_offset = r3_offset.convert(User.from_dict)
    assert len(result3_offset) == 2
    assert result3_offset[0].name == "サンプル太郎"
    assert result3_offset[1].name == "サンプル花子"

    # update
    q4 = QueryBuilder.update(
        target="users",
        query_node=OrNode([
            FieldEquals("name", "サンプル太郎"),
            FieldEquals("name", "サンプル花子"),
        ]),
        override_data={"age": 26},
        return_data=True,
        sort_obj=SingleSort(field="id"),
    ).build()
    r4 = db.execute_query(Query.from_dict(q4.to_dict()))
    assert r4.db_length == 4
    assert r4.update_count == 2
    assert r4.hit_count == 2
    result4 = r4.convert(User.from_dict)
    assert result4[0].name == "サンプル太郎"
    assert result4[0].age == 26
    assert result4[1].name == "サンプル花子"
    assert result4[1].age == 26

    # updateOne
    q6 = QueryBuilder.update_one(
        target="users",
        query_node=FieldEquals("name", "サンプル花子"),
        override_data={"name": "テスト花子"},
        return_data=True,
    ).build()
    r6 = db.execute_query(Query.from_dict(q6.to_dict()))
    assert r6.db_length == 4
    assert r6.update_count == 1
    assert r6.hit_count == 1
    result6 = r6.convert(User.from_dict)
    assert result6[0].name == "テスト花子"
    assert result6[0].age == 26

    # delete
    q8 = QueryBuilder.delete(
        target="users",
        query_node=FieldEquals("name", "テスト花子"),
        sort_obj=SingleSort(field="id"),
        return_data=True,
    ).build()
    r8 = db.execute_query(Query.from_dict(q8.to_dict()))
    assert r8.db_length == 3
    assert r8.update_count == 1
    assert r8.hit_count == 1
    result8 = r8.convert(User.from_dict)
    assert result8[0].name == "テスト花子"
    assert result8[0].age == 26

    # conformToTemplate (User -> User2)
    q10 = QueryBuilder.conform_to_template(
        target="users",
        template=User2(id_="5", name="NoData", age=-1, gender="NoData").to_dict()
    ).build()
    r10 = db.execute_query(Query.from_dict(q10.to_dict()))
    assert r10.db_length == 3

    # renameField (User -> User3)
    q12 = QueryBuilder.rename_field(
        target="users",
        rename_before="id",
        rename_after="serialID",
        return_data=True
    ).build()
    r12 = db.execute_query(Query.from_dict(q12.to_dict()))
    assert r12.db_length == 3
    assert r12.update_count == 3
    assert r12.hit_count == 3
    _ = r12.convert(User3.from_dict)

    # count
    q13 = QueryBuilder.count(target="users").build()
    r13 = db.execute_query(Query.from_dict(q13.to_dict()))
    assert r13.db_length == 3
    assert r13.update_count == 0
    assert r13.hit_count == 3

    # clear
    q14 = QueryBuilder.clear(target="users").build()
    r14 = db.execute_query(Query.from_dict(q14.to_dict()))
    assert r14.db_length == 0
    assert r14.update_count == 3
    assert r14.hit_count == 3


def test_save_and_load():
    now = datetime.now()
    db = DeltaTraceDatabase()
    users = [
        User("1", "サンプル太郎", 25, now, now, {}),
        User("2", "サンプル次郎", 28, now, now, {}),
        User("3", "サンプル三郎", 31, now, now, {}),
        User("4", "サンプル花子", 17, now, now, {}),
    ]
    q = QueryBuilder.add(
        target="users",
        add_data=users,
        cause=Cause(who=Actor(actor_type=EnumActorType.system,
                              actor_id="1",
                              roles=["tester"],
                              permissions=["users:write:all"],
                              collection_permissions={
                                  "users": Permission([EnumQueryType.add]),
                              },
                              context={"otherData": "test"}, ),
                    when=TemporalTrace(),
                    what="The test of serialize and deserialize.",
                    why="test",
                    from_="test",
                    serial="1",
                    chain_parent_serial="1",
                    context={"test": "test"},
                    confidence_score=1.0)
    ).build()

    # JSON 変換テスト
    json_str = json.dumps(q.to_dict())
    assert q.to_dict() == Query.from_dict(json.loads(json_str)).to_dict()

    r1 = db.execute_query(Query.from_dict(q.to_dict()))
    assert r1.is_success

    # シリアライズと復元
    db_dict = db.to_dict()
    db2 = DeltaTraceDatabase.from_dict(db_dict)

    loaded_users = db2.collection("users")
    assert loaded_users.length == 4

    for i in range(len(users)):
        original = User.from_dict(db.collection("users").raw[i])
        loaded = User.from_dict(loaded_users.raw[i])
        assert loaded.id == original.id
        assert loaded.name == original.name
        assert loaded.age == original.age
        assert loaded.created_at.isoformat() == original.created_at.isoformat()


def test_complex_search():
    now = datetime.now()
    db = DeltaTraceDatabase()
    users = [
        User("1", "サンプル太郎", 25, now, now, {"a": "test", "b": 1}),
        User("2", "サンプル次郎", 28, now, now, {"a": "test", "b": 1}),
        User("3", "サンプル三郎", 31, now, now, {"a": "text", "b": 2}),
        User("4", "サンプル花子", 17, now, now, {"a": "text", "b": 3}),
    ]
    q1 = QueryBuilder.add(target="users", add_data=users).build()
    r1 = db.execute_query(q1)
    assert r1.is_success is True

    # ネストオブジェクト検索 (文字列)
    q2 = QueryBuilder.search(
        target="users",
        query_node=FieldEquals("nestedObj.a", "test"),
        sort_obj=SingleSort(field="id", reversed_=False)
    ).build()
    r2 = db.execute_query(Query.from_dict(q2.to_dict()))
    assert r2.hit_count == 2
    result2 = r2.convert(User.from_dict)
    assert result2[0].name == "サンプル太郎"
    assert result2[1].name == "サンプル次郎"

    # ネストオブジェクト検索 (数値)
    q3 = QueryBuilder.search(
        target="users",
        query_node=FieldEquals("nestedObj.b", 1),
        sort_obj=SingleSort(field="id", reversed_=False)
    ).build()
    r3 = db.execute_query(Query.from_dict(q3.to_dict()))
    assert r3.hit_count == 2
    result3 = r3.convert(User.from_dict)
    assert result3[0].name == "サンプル太郎"
    assert result3[1].name == "サンプル次郎"

    # 正規表現検索
    q4 = QueryBuilder.search(
        target="users",
        query_node=FieldEquals("name", "サンプル太郎"),  # 正規表現のダミー
        sort_obj=SingleSort(field="id", reversed_=False)
    ).build()
    r4 = db.execute_query(Query.from_dict(q4.to_dict()))
    assert r4.hit_count == 1


def test_query_serialize():
    q1 = QueryBuilder.search(
        target="users",
        query_node=FieldEquals("nestedObj.a", "test"),
        sort_obj=SingleSort(field="id", reversed_=False)
    ).build()
    q2 = QueryBuilder.search(
        target="users",
        query_node=FieldEquals("nestedObj.a", "test"),
        sort_obj=MultiSort([
            SingleSort(field="id", reversed_=False),
            SingleSort(field="name", reversed_=False),
        ])
    ).build()

    assert isinstance(Query.from_dict(q1.to_dict()).sort_obj, SingleSort)
    assert isinstance(Query.from_dict(q2.to_dict()).sort_obj, MultiSort)


def test_add_with_serial_key():
    # データベース作成とデータ追加
    db = DeltaTraceDatabase()

    # add
    q1 = QueryBuilder.add(
        target="items",
        add_data=[
            Item1(name="itemA"),
            Item1(name="itemB"),
        ],
        serial_key="serialKey",
    ).build()
    r1: QueryResult = db.execute_query(q1)
    assert r1.is_success is True

    # シリアルキーが付与されているかのチェック
    assert db.collection("items").raw[0]["serialKey"] == 0
    assert db.collection("items").raw[1]["serialKey"] == 1

    # シリアライズ・デシリアライズでの復元チェック
    db = DeltaTraceDatabase.from_dict(db.to_dict())

    # add
    q2 = QueryBuilder.add(
        target="items",
        add_data=[
            Item1(name="itemC"),
            Item1(name="itemD"),
        ],
        serial_key="serialKey",
    ).build()
    r2: QueryResult = db.execute_query(q2)
    assert r2.is_success is True

    # シリアルキーが付与されているかのチェック
    assert db.collection("items").raw[2]["serialKey"] == 2
    assert db.collection("items").raw[3]["serialKey"] == 3

    # clear add
    q3 = QueryBuilder.clear_add(
        target="items",
        add_data=[
            Item1(name="itemC"),
            Item1(name="itemD"),
        ],
        serial_key="serialKey",
    ).build()
    r3: QueryResult = db.execute_query(q3)
    assert r3.is_success is True

    # シリアルキーが付与されているかのチェック
    assert db.collection("items").raw[0]["serialKey"] == 4
    assert db.collection("items").raw[1]["serialKey"] == 5

    # clear add with reset serial
    q4 = QueryBuilder.clear_add(
        target="items",
        add_data=[
            Item1(name="itemC"),
            Item1(name="itemD"),
        ],
        serial_key="serialKey",
        reset_serial=True
    ).build()
    r4: QueryResult = db.execute_query(q4)
    assert r4.is_success is True

    # シリアルキーがリセット後に付与されているかのチェック
    assert db.collection("items").raw[0]["serialKey"] == 0
    assert db.collection("items").raw[1]["serialKey"] == 1


def test_add_with_return_data():
    # データベース作成
    db = DeltaTraceDatabase()

    # データ追加クエリ作成
    q1 = QueryBuilder.add(
        target='items',
        add_data=[
            Item1(name="itemA"),
            Item1(name="itemB")
        ],
        serial_key="serialKey",
        return_data=True
    ).build()

    # クエリ実行
    r1 = db.execute_query(q1)

    # 結果の検証
    assert r1.is_success is True
    assert len(r1.result) > 0

    r_items = r1.convert(Item1.from_dict)

    assert r_items[0].serial_key == 0
    assert r_items[1].serial_key == 1



# coding: utf-8
import json

from delta_trace_db.query.nodes.comparison_node import (
    FieldEquals,
    FieldNotEquals,
    FieldGreaterThan,
    FieldGreaterThanOrEqual,
    FieldLessThan,
    FieldLessThanOrEqual,
)
from delta_trace_db.query.nodes.logical_node import AndNode, OrNode, NotNode
from delta_trace_db.query.nodes.query_node import QueryNode


# --------------------------
# Comparison Nodes
# --------------------------

def test_field_equals():
    node = FieldEquals('name', 'Alice')
    assert node.evaluate({'name': 'Alice'})
    assert not node.evaluate({'name': 'Bob'})

def test_field_not_equals():
    node = FieldNotEquals('age', 30)
    assert node.evaluate({'age': 25})
    assert not node.evaluate({'age': 30})

def test_field_greater_than():
    node = FieldGreaterThan('score', 80)
    assert node.evaluate({'score': 90})
    assert not node.evaluate({'score': 70})

def test_field_greater_than_or_equal():
    node = FieldGreaterThanOrEqual('score', 80)
    assert node.evaluate({'score': 80})
    assert not node.evaluate({'score': 79})

def test_field_less_than():
    node = FieldLessThan('score', 80)
    assert node.evaluate({'score': 70})
    assert not node.evaluate({'score': 90})

def test_field_less_than_or_equal():
    node = FieldLessThanOrEqual('score', 80)
    assert node.evaluate({'score': 80})
    assert not node.evaluate({'score': 81})


# --------------------------
# Logical Nodes
# --------------------------

def test_and_node():
    node = AndNode([FieldEquals('x', 1), FieldEquals('y', 2)])
    assert node.evaluate({'x': 1, 'y': 2})
    assert not node.evaluate({'x': 1, 'y': 3})

def test_or_node():
    node = OrNode([FieldEquals('x', 1), FieldEquals('y', 2)])
    assert node.evaluate({'x': 0, 'y': 2})
    assert not node.evaluate({'x': 0, 'y': 0})

def test_not_node():
    node = NotNode(FieldEquals('z', 5))
    assert not node.evaluate({'z': 5})
    assert node.evaluate({'z': 4})


# --------------------------
# Query Nodes
# --------------------------

def test_query_node_mixed():
    query = AndNode([
        FieldEquals('type', 'user'),
        OrNode([FieldGreaterThan('age', 18), FieldEquals('role', 'admin')]),
    ])
    assert query.evaluate({'type': 'user', 'age': 20})
    assert query.evaluate({'type': 'user', 'role': 'admin'})
    assert not query.evaluate({'type': 'user', 'age': 10})
    assert not query.evaluate({'type': 'bot', 'age': 30})

def test_nested_logical():
    query = NotNode(
        AndNode([FieldEquals('active', True), FieldLessThan('count', 5)])
    )
    assert not query.evaluate({'active': True, 'count': 3})
    assert query.evaluate({'active': False, 'count': 3})
    assert query.evaluate({'active': True, 'count': 6})

def test_serialization_field_equals():
    node = FieldEquals('name', 'Alice')
    json_str = json.dumps(node.to_dict())
    restored = FieldEquals.from_dict(json.loads(json_str))
    assert restored.evaluate({'name': 'Alice'})
    assert not restored.evaluate({'name': 'Bob'})

def test_serialization_field_greater_than_or_equal():
    node = FieldGreaterThanOrEqual('score', 50)
    json_str = json.dumps(node.to_dict())
    restored = FieldGreaterThanOrEqual.from_dict(json.loads(json_str))
    assert restored.evaluate({'score': 60})
    assert not restored.evaluate({'score': 40})

def test_serialization_and_node():
    node = AndNode([FieldEquals('type', 'user'), FieldGreaterThan('age', 18)])
    json_str = json.dumps(node.to_dict())
    restored = AndNode.from_dict(json.loads(json_str))
    assert restored.evaluate({'type': 'user', 'age': 20})
    assert not restored.evaluate({'type': 'user', 'age': 10})

def test_serialization_not_or_node():
    node = NotNode(OrNode([FieldEquals('x', 1), FieldEquals('y', 2)]))
    json_str = json.dumps(node.to_dict())
    restored = NotNode.from_dict(json.loads(json_str))
    assert restored.evaluate({'x': 0, 'y': 0})
    assert not restored.evaluate({'x': 1, 'y': 0})

def test_serialization_complex_query_node():
    original = AndNode([
        FieldEquals('role', 'admin'),
        OrNode([FieldLessThan('logins', 5), FieldGreaterThan('age', 40)]),
    ])
    json_str = json.dumps(original.to_dict())
    decoded = json.loads(json_str)
    restored = AndNode.from_dict(decoded)
    assert restored.evaluate({'role': 'admin', 'logins': 2})
    assert restored.evaluate({'role': 'admin', 'age': 45})
    assert not restored.evaluate({'role': 'admin', 'logins': 10, 'age': 30})
    assert not restored.evaluate({'role': 'user', 'logins': 2})

def test_from_dict_field_equals():
    original = FieldEquals('foo', 'bar')
    json_str = json.dumps(original.to_dict())
    restored = QueryNode.from_dict(json.loads(json_str))
    assert isinstance(restored, FieldEquals)
    assert restored.evaluate({'foo': 'bar'})

def test_from_dict_nested_and_node():
    original = AndNode([FieldEquals('a', 1), FieldGreaterThan('b', 10)])
    json_str = json.dumps(original.to_dict())
    restored = QueryNode.from_dict(json.loads(json_str))
    assert isinstance(restored, AndNode)
    assert restored.evaluate({'a': 1, 'b': 11})
    assert not restored.evaluate({'a': 1, 'b': 5})

def test_from_dict_complex_not_or_node():
    original = NotNode(OrNode([FieldEquals('x', True), FieldLessThan('y', 5)]))
    json_str = json.dumps(original.to_dict())
    restored = QueryNode.from_dict(json.loads(json_str))
    assert isinstance(restored, NotNode)
    assert restored.evaluate({'x': False, 'y': 10})  # NOT(FALSE OR FALSE) = TRUE
    assert not restored.evaluate({'x': True, 'y': 10})  # NOT(TRUE OR FALSE) = FALSE



# coding: utf-8
import functools
from typing import Any, Callable, Dict, List, Set
from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.db.util_copy import UtilCopy
from delta_trace_db.query.nodes.query_node import QueryNode
from delta_trace_db.query.query import Query
from delta_trace_db.query.query_result import QueryResult


class Collection(CloneableFile):
    class_name = "Collection"
    version = "11"

    def __init__(self):
        super().__init__()
        self._data: List[Dict[str, Any]] = []
        self._serial_num: int = 0
        self.listeners: Set[Callable[[], None]] = set()
        self._is_transaction_mode: bool = False
        self.run_notify_listeners_in_transaction: bool = False

    def change_transaction_mode(self, is_transaction_mode: bool):
        self._is_transaction_mode = is_transaction_mode
        self.run_notify_listeners_in_transaction = False

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "Collection":
        instance = cls()
        instance._data = UtilCopy.jsonable_deep_copy(src.get("data", []))
        instance._serial_num = src.get("serialNum", 0)
        return instance

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.class_name,
            "version": self.version,
            "data": UtilCopy.jsonable_deep_copy(self._data),
            "serialNum": self._serial_num
        }

    def clone(self) -> "Collection":
        return Collection.from_dict(self.to_dict())

    @property
    def raw(self) -> List[Dict[str, Any]]:
        return self._data

    @property
    def length(self) -> int:
        return len(self._data)

    def add_listener(self, cb: Callable[[], None]):
        self.listeners.add(cb)

    def remove_listener(self, cb: Callable[[], None]):
        self.listeners.discard(cb)

    def notify_listeners(self):
        if not self._is_transaction_mode:
            for cb in self.listeners:
                cb()
        else:
            self.run_notify_listeners_in_transaction = True

    def _evaluate(self, record: Dict[str, Any], node: QueryNode) -> bool:
        return node.evaluate(record)

    def add_all(self, q: Query) -> QueryResult:
        add_data = UtilCopy.jsonable_deep_copy(q.add_data)
        added_items = []
        if q.serial_key is not None:
            # 対象キーの存在チェック
            for item in add_data:
                if q.serial_key not in item:
                    return QueryResult(
                        is_success=False,
                        target=q.target,
                        type_=q.type,
                        result=[],
                        db_length=len(self._data),
                        update_count=0,
                        hit_count=0,
                        error_message=f'The target serialKey does not exist. serialKey:{q.serial_key}',
                    )

            for item in add_data:
                serial_num = self._serial_num
                item[q.serial_key] = serial_num
                self._serial_num += 1
                self._data.append(item)
                if q.return_data:
                    added_items.append(item)
        else:
            self._data.extend(add_data)
            if q.return_data:
                added_items.extend(add_data)
        self.notify_listeners()
        return QueryResult(True, q.target, q.type, UtilCopy.jsonable_deep_copy(added_items), len(self._data),
                           len(add_data), 0)

    def update(self, q: Query, is_single_target: bool) -> QueryResult:
        if q.return_data:
            r = []
            for item in self._data:
                if self._evaluate(item, q.query_node):
                    item.update(UtilCopy.jsonable_deep_copy(q.override_data))
                    r.append(item)
                    if is_single_target:
                        break
            if r:
                # 要素が空ではないなら通知を発行。
                self.notify_listeners()
            return QueryResult(True, q.target, q.type, UtilCopy.jsonable_deep_copy(r), len(self._data), len(r), len(r))
        else:
            count = 0
            for item in self._data:
                if self._evaluate(item, q.query_node):
                    item.update(UtilCopy.jsonable_deep_copy(q.override_data))
                    count += 1
                    if is_single_target:
                        break
            if count > 0:
                self.notify_listeners()
            return QueryResult(True, q.target, q.type, [], len(self._data), count, count)

    def delete(self, q: Query) -> QueryResult:
        if q.return_data:
            deleted_items = [item for item in self._data if self._evaluate(item, q.query_node)]
            self._data = [item for item in self._data if not self._evaluate(item, q.query_node)]
            if deleted_items:
                self.notify_listeners()
            return QueryResult(True, q.target, q.type, UtilCopy.jsonable_deep_copy(deleted_items), len(self._data),
                               len(deleted_items),
                               len(deleted_items))
        else:
            count = sum(1 for item in self._data if self._evaluate(item, q.query_node))
            self._data = [item for item in self._data if not self._evaluate(item, q.query_node)]
            if count > 0:
                self.notify_listeners()
            return QueryResult(True, q.target, q.type, [], len(self._data), count, count)

    def delete_one(self, q: Query) -> QueryResult:
        deleted_items = []
        for i, item in enumerate(self._data):
            if self._evaluate(item, q.query_node):
                deleted_items.append(item)
                del self._data[i]
                break
        if deleted_items:
            self.notify_listeners()
        return QueryResult(True, q.target, q.type, UtilCopy.jsonable_deep_copy(deleted_items), len(self._data),
                           len(deleted_items),
                           len(deleted_items))

    def search(self, q: Query) -> QueryResult:
        r: List[Dict[str, Any]] = []
        # 検索
        for item in self._data:
            if self._evaluate(item, q.query_node):
                r.append(item)
        hit_count = len(r)
        # ソートやページングのオプション
        if q.sort_obj is not None:
            sorted_list = list(r)
            sorted_list.sort(key=functools.cmp_to_key(q.sort_obj.get_comparator()))
            r = sorted_list
            if q.offset is not None:
                if q.offset > 0:
                    r = r[q.offset:]
            else:
                if q.start_after is not None:
                    try:
                        index = r.index(q.start_after)
                        if index != -1 and index + 1 < len(r):
                            r = r[index + 1:]
                        elif index != -1 and index + 1 >= len(r):
                            r = []
                    except ValueError:
                        pass
                elif q.end_before is not None:
                    try:
                        index = r.index(q.end_before)
                        if index != -1:
                            r = r[:index]
                    except ValueError:
                        pass
        if q.limit is not None:
            if q.offset is None and q.start_after is None and q.end_before is not None:
                r = r[-q.limit:] if len(r) > q.limit else r
            else:
                r = r[:q.limit]
        return QueryResult(
            is_success=True,
            target=q.target,
            type_=q.type,
            result=UtilCopy.jsonable_deep_copy(r),
            db_length=len(self._data),
            update_count=0,
            hit_count=hit_count,
        )

    def get_all(self, q: Query) -> QueryResult:
        r = UtilCopy.jsonable_deep_copy(self._data)
        if q.sort_obj:
            r.sort(key=q.sort_obj.get_comparator())
        return QueryResult(True, q.target, q.type, r, len(self._data), 0, len(r))

    def conform_to_template(self, q: Query) -> QueryResult:
        for item in self._data:
            keys_to_remove = [k for k in item.keys() if k not in q.template]
            for k in keys_to_remove:
                item.pop(k)
            for k, v in q.template.items():
                if k not in item:
                    item[k] = UtilCopy.jsonable_deep_copy(v)
        self.notify_listeners()
        return QueryResult(True, q.target, q.type, [], len(self._data), len(self._data), len(self._data))

    def rename_field(self, q: Query) -> QueryResult:
        r = []
        for item in self._data:
            if q.rename_before not in item:
                return QueryResult(False, q.target, q.type, [], len(self._data), 0, 0,
                                   f'The target key does not exist. key:{q.rename_before}')
            if q.rename_after in item:
                return QueryResult(False, q.target, q.type, [], len(self._data), 0, 0,
                                   f'An existing key was specified as the new key. key:{q.rename_after}')
        update_count = 0
        for item in self._data:
            item[q.rename_after] = item[q.rename_before]
            del item[q.rename_before]
            update_count += 1
            if q.return_data:
                r.append(item)
        self.notify_listeners()
        return QueryResult(True, q.target, q.type, UtilCopy.jsonable_deep_copy(r), len(self._data), update_count,
                           update_count)

    def count(self, q: Query) -> QueryResult:
        return QueryResult(True, q.target, q.type, [], len(self._data), 0, len(self._data))

    def clear(self, q: Query) -> QueryResult:
        pre_len = len(self._data)
        self._data.clear()
        if q.reset_serial:
            self._serial_num = 0
        self.notify_listeners()
        return QueryResult(True, q.target, q.type, [], 0, pre_len, pre_len)

    def clear_add(self, q: Query) -> QueryResult:
        pre_len = len(self._data)
        self._data.clear()
        if q.reset_serial:
            self._serial_num = 0
        added_items = []
        add_data = UtilCopy.jsonable_deep_copy(q.add_data)
        if q.serial_key is not None:
            # 対象キーの存在チェック
            for item in add_data:
                if q.serial_key not in item:
                    return QueryResult(
                        is_success=False,
                        target=q.target,
                        type_=q.type,
                        result=[],
                        db_length=len(self._data),
                        update_count=0,
                        hit_count=0,
                        error_message=f'The target serialKey does not exist. serialKey:{q.serial_key}',
                    )

            for item in add_data:
                serial_num = self._serial_num
                item[q.serial_key] = serial_num
                self._serial_num += 1
                self._data.append(item)
                if q.return_data:
                    added_items.append(item)
        else:
            self._data.extend(add_data)
            if q.return_data:
                added_items.extend(add_data)
        self.notify_listeners()
        return QueryResult(True, q.target, q.type, UtilCopy.jsonable_deep_copy(added_items), len(self._data), pre_len,
                           pre_len)



# coding: utf-8
from typing import List, Dict, Any

from file_state_manager.cloneable_file import CloneableFile

from delta_trace_db.query.query import Query


class TransactionQuery(CloneableFile):
    """
    (en) This is a query class that supports transaction processing.
    It internally stores normal query classes as an array,
    and the targets are processed as transactions.

    (ja) トランザクション処理に対応したクエリクラスです。
    内部に通常のクエリクラスを配列で保持しており、対象はトランザクション処理されます。
    """
    className: str = "TransactionQuery"
    version: str = "1"

    def __init__(self, queries: List[Query]):
        """
        :param queries: The transaction targets.
        """
        super().__init__()
        self.queries = queries

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "TransactionQuery":
        """
        (en) Restore this object from the dictionary.
        (ja) このオブジェクトを辞書から復元します。

        :param src: A dictionary made with to_dict of this class.
        """
        q = [Query.from_dict(i) for i in src["queries"]]
        return cls(queries=q)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.className,
            "version": self.version,
            "queries": [i.to_dict() for i in self.queries]
        }

    def clone(self) -> "TransactionQuery":
        return TransactionQuery.from_dict(self.to_dict())



# coding: utf-8
from typing import Any, Callable, Dict, List
from delta_trace_db.query.sort.abstract_sort import AbstractSort
from delta_trace_db.query.sort.single_sort import SingleSort

class MultiSort(AbstractSort):
    class_name = "MultiSort"
    version = "1"

    def __init__(self, sort_orders: List[SingleSort]):
        self.sort_orders = sort_orders

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "MultiSort":
        orders = [SingleSort.from_dict(e) for e in src.get("sortOrders", [])]
        return cls(orders)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.class_name,
            "version": self.version,
            "sortOrders": [s.to_dict() for s in self.sort_orders],
        }

    def get_comparator(self) -> Callable[[Dict[str, Any], Dict[str, Any]], int]:
        def comparator(a: Dict[str, Any], b: Dict[str, Any]) -> int:
            for sort_obj in self.sort_orders:
                comp = sort_obj.get_comparator()(a, b)
                if comp != 0:
                    return comp
            return 0
        return comparator



# coding: utf-8
from datetime import datetime, timedelta
from file_state_manager.cloneable_file import CloneableFile

from delta_trace_db.db.delta_trace_db_core import DeltaTraceDatabase
from delta_trace_db.query.nodes.comparison_node import FieldEquals
from delta_trace_db.query.query_builder import QueryBuilder
from delta_trace_db.query.transaction_query import TransactionQuery


class User(CloneableFile):
    def __init__(self, id_: str, name: str, age: int, created_at: datetime, updated_at: datetime, nested_obj: dict):
        super().__init__()
        self.id = id_
        self.name = name
        self.age = age
        self.created_at = created_at
        self.updated_at = updated_at
        self.nested_obj = nested_obj

    @classmethod
    def from_dict(cls, src: dict) -> "User":
        return User(
            id_=src["id"],
            name=src["name"],
            age=src["age"],
            created_at=datetime.fromisoformat(src["createdAt"]),
            updated_at=datetime.fromisoformat(src["updatedAt"]),
            nested_obj=src["nestedObj"],
        )

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
            "age": self.age,
            "createdAt": self.created_at.isoformat(),
            "updatedAt": datetime.now().isoformat(),
            "nestedObj": dict(self.nested_obj),
        }

    def clone(self) -> "User":
        return User.from_dict(self.to_dict())


def test_transaction_query():
    now = datetime.now()
    db = DeltaTraceDatabase()
    users = [
        User(id_="1", name="サンプル太郎", age=25, created_at=now + timedelta(days=0),
             updated_at=now + timedelta(days=0), nested_obj={}),
        User(id_="2", name="サンプル次郎", age=28, created_at=now + timedelta(days=1),
             updated_at=now + timedelta(days=1), nested_obj={}),
        User(id_="3", name="サンプル三郎", age=31, created_at=now + timedelta(days=2),
             updated_at=now + timedelta(days=2), nested_obj={}),
        User(id_="4", name="サンプル花子", age=17, created_at=now + timedelta(days=3),
             updated_at=now + timedelta(days=3), nested_obj={}),
    ]

    # add
    q1 = QueryBuilder.add(target="users1", add_data=users).build()
    q2 = QueryBuilder.add(target="users2", add_data=users).build()
    db.execute_query(q1)
    db.execute_query(q2)

    # Failed transaction
    tq1 = TransactionQuery(
        queries=[
            QueryBuilder.update(
                target="users1",
                query_node=FieldEquals("id", 3),  # type error
                override_data={"id": 5},
                return_data=True,
                must_affect_at_least_one=True,
            ).build(),
            QueryBuilder.clear(target="users2").build(),
        ]
    )
    result = db.execute_query_object(tq1)
    assert not result.is_success
    assert db.collection("users2").length != 0

    # Success transaction
    tq2 = TransactionQuery(
        queries=[
            QueryBuilder.update(
                target="users1",
                query_node=FieldEquals("id", "3"),
                override_data={"id": "5"},
                return_data=True,
                must_affect_at_least_one=True,
            ).build(),
            QueryBuilder.clear(target="users2").build(),
        ]
    )
    result2 = db.execute_query_object(tq2)
    assert result2.is_success
    assert db.collection("users2").length == 0



# coding: utf-8
from abc import ABC
from typing import Dict, Any

from file_state_manager.cloneable_file import CloneableFile


class QueryExecutionResult(CloneableFile, ABC):
    def __init__(self, is_success: bool):
        super().__init__()
        self.is_success: bool = is_success

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "QueryExecutionResult":
        # 遅延インポート
        from delta_trace_db.query.query_result import QueryResult
        from delta_trace_db.query.transaction_query_result import TransactionQueryResult
        class_name = src.get("className")
        if class_name == "QueryResult":
            return QueryResult.from_dict(src)
        elif class_name == "TransactionQueryResult":
            return TransactionQueryResult.from_dict(src)
        else:
            raise ValueError("QueryExecutionResult: The object cannot be converted.")



import pytest
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.query_result import QueryResult


def test_version_up():
    try:
        qr = QueryResult.from_dict({
            "className": "QueryResult",
            "version": "5",
            "isSuccess": True,
            # target が無しでも復元できるか。互換性の確認。
            # "target": "abc",
            "type": EnumQueryType.add.name,
            "result": [],
            "dbLength": 0,
            "updateCount": 0,
            "hitCount": 0,
            "errorMessage": None,
        })
        assert qr.target == ""
    except Exception as e:
        # 例外が出たらテスト失敗
        pytest.fail(f"Unexpected exception: {e}")



# coding: utf-8
from typing import Optional, Dict, Any
from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.db.util_copy import UtilCopy
from delta_trace_db.query.cause.actor import Actor
from delta_trace_db.query.cause.temporal_trace.temporal_trace import TemporalTrace


class Cause(CloneableFile):
    class_name = "Cause"
    version = "1"

    def __init__(self, who: Actor, when: TemporalTrace, what: str, why: str, from_: str, serial: Optional[str] = None,
                 chain_parent_serial: Optional[str] = None, context: Optional[Dict[str, Any]] = None,
                 confidence_score: float = 1.0):
        super().__init__()
        self.serial = serial
        self.chain_parent_serial = chain_parent_serial
        self.who = who
        self.when = when
        self.what = what
        self.why = why
        self.from_ = from_
        self.context = context
        self.confidence_score = confidence_score

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "Cause":
        return cls(
            serial=src.get("serial"),
            chain_parent_serial=src.get("chainParentSerial"),
            who=Actor.from_dict(src["who"]),
            when=TemporalTrace.from_dict(src["when"]),
            what=src["what"],
            why=src["why"],
            from_=src["from"],
            context=src.get("context"),
            confidence_score=src.get("confidenceScore", 1.0),
        )

    def clone(self) -> "Cause":
        return Cause.from_dict(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.class_name,
            "version": self.version,
            "serial": self.serial,
            "chainParentSerial": self.chain_parent_serial,
            "who": self.who.to_dict(),
            "when": self.when.to_dict(),
            "what": self.what,
            "why": self.why,
            "from": self.from_,
            "context": UtilCopy.jsonable_deep_copy(self.context),
            "confidenceScore": self.confidence_score,
        }



# coding: utf-8
from typing import List, Dict, Any, Callable

from delta_trace_db.db.util_copy import UtilCopy
from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.query_execution_result import QueryExecutionResult

class QueryResult(QueryExecutionResult):
    class_name: str = "QueryResult"
    version: str = "5"

    def __init__(
        self,
        is_success: bool,
        target: str,
        type_: EnumQueryType,
        result: List[Dict[str, Any]],
        db_length: int,
        update_count: int,
        hit_count: int,
        error_message: str | None = None,
    ):
        super().__init__(is_success=is_success)
        self.target: str = target
        self.type: EnumQueryType = type_
        self.result: List[Dict[str, Any]] = result
        self.db_length: int = db_length
        self.update_count: int = update_count
        self.hit_count: int = hit_count
        self.error_message: str | None = error_message

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "QueryResult":
        return cls(
            is_success=src["isSuccess"],
            target=src.get("target", ""),
            type_=EnumQueryType[src["type"]],
            result=list(src["result"]),
            db_length=src["dbLength"],
            update_count=src["updateCount"],
            hit_count=src["hitCount"],
            error_message=src.get("errorMessage"),
        )

    def convert(self, from_dict: Callable) -> List:
        return [from_dict(i) for i in self.result]

    def clone(self) -> "QueryResult":
        return self.from_dict(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        return {
            "className": self.class_name,
            "version": self.version,
            "isSuccess": self.is_success,
            "target": self.target,
            "type": self.type.name,
            "result": UtilCopy.jsonable_deep_copy(self.result),
            "dbLength": self.db_length,
            "updateCount": self.update_count,
            "hitCount": self.hit_count,
            "errorMessage": self.error_message,
        }



# coding: utf-8
from typing import Any, Dict, List, Optional

from delta_trace_db.query.enum_query_type import EnumQueryType
from delta_trace_db.query.query_builder import QueryBuilder
from delta_trace_db.query.query import Query
from delta_trace_db.query.nodes.query_node import QueryNode


class RawQueryBuilder(QueryBuilder):
    def __init__(
            self,
            target: str,
            type_: EnumQueryType,
            raw_add_data: Optional[List[Dict[str, Any]]] = None,
            raw_template: Optional[Dict[str, Any]] = None,
            query_node: Optional[QueryNode] = None,
            override_data: Optional[Dict[str, Any]] = None,
            return_data: Optional[bool] = None,
            sort_obj: Optional[Any] = None,
            offset: Optional[int] = None,
            start_after: Optional[Any] = None,
            end_before: Optional[Any] = None,
            rename_before: Optional[str] = None,
            rename_after: Optional[str] = None,
            limit: Optional[int] = None,
            must_affect_at_least_one: bool = True,
            serial_key: Optional[str] = None,
            reset_serial: bool = False,
            cause: Optional[Any] = None,
    ):
        super().__init__(
            target=target,
            type_=type_,
            query_node=query_node,
            override_data=override_data,
            return_data=return_data,
            sort_obj=sort_obj,
            offset=offset,
            start_after=start_after,
            end_before=end_before,
            rename_before=rename_before,
            rename_after=rename_after,
            limit=limit,
            must_affect_at_least_one=must_affect_at_least_one,
            serial_key=serial_key,
            reset_serial=reset_serial,
            cause=cause
        )
        self.raw_add_data = raw_add_data
        self.raw_template = raw_template

    @classmethod
    def add(cls, target: str, raw_add_data: List[Dict[str, Any]],
            return_data: bool = False,
            must_affect_at_least_one: bool = True,
            serial_key: Optional[str] = None,
            cause: Optional[Any] = None):
        return cls(target=target, type_=EnumQueryType.add, raw_add_data=raw_add_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   serial_key=serial_key,
                   cause=cause)

    @classmethod
    def update(
            cls,
            target: str,
            query_node: QueryNode,
            override_data: Optional[Dict[str, Any]],
            return_data: bool = False,
            sort_obj: Optional[Any] = None,
            must_affect_at_least_one: bool = True,
            cause: Optional[Any] = None
    ):
        return cls(
            target=target,
            type_=EnumQueryType.update,
            query_node=query_node,
            override_data=override_data,
            return_data=return_data,
            sort_obj=sort_obj,
            must_affect_at_least_one=must_affect_at_least_one,
            cause=cause
        )

    @classmethod
    def update_one(
            cls,
            target: str,
            query_node: QueryNode,
            override_data: Optional[Dict[str, Any]],
            return_data: bool = False,
            must_affect_at_least_one: bool = True,
            cause: Optional[Any] = None
    ):
        return cls(
            target=target,
            type_=EnumQueryType.updateOne,
            query_node=query_node,
            override_data=override_data,
            return_data=return_data,
            must_affect_at_least_one=must_affect_at_least_one,
            cause=cause
        )

    @classmethod
    def delete(
            cls,
            target: str,
            query_node: QueryNode,
            return_data: bool = False,
            sort_obj: Optional[Any] = None,
            must_affect_at_least_one: bool = True,
            cause: Optional[Any] = None
    ):
        return cls(
            target=target,
            type_=EnumQueryType.delete,
            query_node=query_node,
            return_data=return_data,
            sort_obj=sort_obj,
            must_affect_at_least_one=must_affect_at_least_one,
            cause=cause
        )

    @classmethod
    def delete_one(
            cls,
            target: str,
            query_node: QueryNode,
            return_data: bool = False,
            must_affect_at_least_one: bool = True,
            cause: Optional[Any] = None
    ):
        return cls(
            target=target,
            type_=EnumQueryType.deleteOne,
            query_node=query_node,
            return_data=return_data,
            must_affect_at_least_one=must_affect_at_least_one,
            cause=cause
        )

    @classmethod
    def search(
            cls,
            target: str,
            query_node: QueryNode,
            sort_obj: Optional[Any] = None,
            offset: Optional[int] = None,
            start_after: Optional[Any] = None,
            end_before: Optional[Any] = None,
            limit: Optional[int] = None,
            cause: Optional[Any] = None
    ):
        return cls(
            target=target,
            type_=EnumQueryType.search,
            query_node=query_node,
            sort_obj=sort_obj,
            offset=offset,
            start_after=start_after,
            end_before=end_before,
            limit=limit,
            cause=cause
        )

    @classmethod
    def get_all(cls, target: str, sort_obj: Optional[Any] = None, cause: Optional[Any] = None):
        return cls(target=target, type_=EnumQueryType.getAll, sort_obj=sort_obj, cause=cause)

    @classmethod
    def conform_to_template(
            cls,
            target: str,
            raw_template: Optional[Dict[str, Any]] = None,
            must_affect_at_least_one: bool = True,
            cause: Optional[Any] = None
    ):
        return cls(
            target=target,
            type_=EnumQueryType.conformToTemplate,
            raw_template=raw_template,
            must_affect_at_least_one=must_affect_at_least_one,
            cause=cause
        )

    @classmethod
    def rename_field(
            cls,
            target: str,
            rename_before: str,
            rename_after: str,
            return_data: bool = False,
            must_affect_at_least_one: bool = True,
            cause: Optional[Any] = None
    ):
        return cls(
            target=target,
            type_=EnumQueryType.renameField,
            rename_before=rename_before,
            rename_after=rename_after,
            return_data=return_data,
            must_affect_at_least_one=must_affect_at_least_one,
            cause=cause
        )

    @classmethod
    def count(cls, target: str, cause: Optional[Any] = None):
        return cls(target=target, type_=EnumQueryType.count, cause=cause)

    @classmethod
    def clear(cls, target: str, must_affect_at_least_one: bool = True, reset_serial: bool = False,
              cause: Optional[Any] = None):
        return cls(target=target, type_=EnumQueryType.clear, must_affect_at_least_one=must_affect_at_least_one,
                   reset_serial=reset_serial,
                   cause=cause)

    @classmethod
    def clear_add(cls, target: str, raw_add_data: List[Dict[str, Any]],
                  return_data: bool = False,
                  must_affect_at_least_one: bool = True,
                  serial_key: Optional[str] = None,
                  reset_serial: bool = False,
                  cause: Optional[Any] = None):
        return cls(target=target, type_=EnumQueryType.clearAdd, raw_add_data=raw_add_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   serial_key=serial_key,
                   reset_serial=reset_serial,
                   cause=cause)

    def build(self) -> Query:
        return Query(
            target=self.target,
            type_=self.type,
            add_data=self.raw_add_data,
            override_data=self.override_data,
            template=self.raw_template,
            query_node=self.query_node,
            sort_obj=self.sort_obj,
            offset=self.offset,
            start_after=self.start_after,
            end_before=self.end_before,
            rename_before=self.rename_before,
            rename_after=self.rename_after,
            limit=self.limit,
            return_data=self.return_data,
            must_affect_at_least_one=self.must_affect_at_least_one,
            serial_key=self.serial_key,
            reset_serial=self.reset_serial,
            cause=self.cause
        )



# coding: utf-8
from enum import Enum

class EnumQueryType(Enum):
    """
    (en) An enum that defines the type of query.

    (ja) クエリの種類を定義したEnum。
    """
    add = "add"
    update = "update"
    updateOne = "updateOne"
    delete = "delete"
    deleteOne = "deleteOne"
    search = "search"
    getAll = "getAll"
    conformToTemplate = "conformToTemplate"  # DB shape change.
    renameField = "renameField"  # DB field name change.
    count = "count"  # get all items count.
    clear = "clear"  # delete all items.
    clearAdd = "clearAdd"  # clear then add.



# coding: utf-8
from typing import List, Dict, Optional, Any

from .cause.cause import Cause
from .enum_query_type import EnumQueryType
from .query import Query
from file_state_manager.cloneable_file import CloneableFile
from delta_trace_db.query.nodes.query_node import QueryNode
from .sort.abstract_sort import AbstractSort


class QueryBuilder:
    def __init__(self,
                 target: str,
                 type_: EnumQueryType,
                 add_data: Optional[List[CloneableFile]] = None,
                 override_data: Optional[Dict[str, Any]] = None,
                 template: Optional[Dict[str, Any]] = None,
                 query_node: Optional[QueryNode] = None,
                 sort_obj: Optional[AbstractSort] = None,
                 offset: Optional[int] = None,
                 start_after: Optional[Dict[str, Any]] = None,
                 end_before: Optional[Dict[str, Any]] = None,
                 rename_before: Optional[str] = None,
                 rename_after: Optional[str] = None,
                 limit: Optional[int] = None,
                 return_data: bool = False,
                 must_affect_at_least_one: bool = True,
                 serial_key: Optional[str] = None,
                 reset_serial: bool = False,
                 cause: Optional[Cause] = None):
        self.target = target
        self.type = type_
        self.add_data = add_data
        self.override_data = override_data
        self.template = template
        self.query_node = query_node
        self.sort_obj = sort_obj
        self.offset = offset
        self.start_after = start_after
        self.end_before = end_before
        self.rename_before = rename_before
        self.rename_after = rename_after
        self.limit = limit
        self.return_data = return_data
        self.must_affect_at_least_one = must_affect_at_least_one
        self.serial_key = serial_key
        self.reset_serial = reset_serial
        self.cause = cause

    @classmethod
    def add(cls, target: str,
            add_data: List[CloneableFile],
            return_data: bool = False,
            must_affect_at_least_one: bool = True,
            serial_key: Optional[str] = None,
            cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.add,
                   add_data=add_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   serial_key=serial_key,
                   cause=cause)

    @classmethod
    def update(cls, target: str,
               query_node: QueryNode,
               override_data: Dict[str, Any],
               return_data: bool = False,
               sort_obj: Optional[AbstractSort] = None,
               must_affect_at_least_one: bool = True,
               cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.update,
                   query_node=query_node,
                   override_data=override_data,
                   return_data=return_data,
                   sort_obj=sort_obj,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def update_one(cls, target: str,
                   query_node: QueryNode,
                   override_data: Dict[str, Any],
                   return_data: bool = False,
                   must_affect_at_least_one: bool = True,
                   cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.updateOne,
                   query_node=query_node,
                   override_data=override_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def delete(cls, target: str,
               query_node: QueryNode,
               return_data: bool = False,
               sort_obj: Optional[AbstractSort] = None,
               must_affect_at_least_one: bool = True,
               cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.delete,
                   query_node=query_node,
                   return_data=return_data,
                   sort_obj=sort_obj,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def delete_one(cls, target: str,
                   query_node: QueryNode,
                   return_data: bool = False,
                   must_affect_at_least_one: bool = True,
                   cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.deleteOne,
                   query_node=query_node,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def search(cls, target: str,
               query_node: QueryNode,
               sort_obj: Optional[AbstractSort] = None,
               offset: Optional[int] = None,
               start_after: Optional[Dict[str, Any]] = None,
               end_before: Optional[Dict[str, Any]] = None,
               limit: Optional[int] = None,
               cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.search,
                   query_node=query_node,
                   sort_obj=sort_obj,
                   offset=offset,
                   start_after=start_after,
                   end_before=end_before,
                   limit=limit,
                   cause=cause)

    @classmethod
    def get_all(cls, target: str,
                sort_obj: Optional[AbstractSort] = None,
                cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.getAll,
                   sort_obj=sort_obj,
                   cause=cause)

    @classmethod
    def conform_to_template(cls, target: str,
                            template: Dict[str, Any],
                            must_affect_at_least_one: bool = True,
                            cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.conformToTemplate,
                   template=template,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def rename_field(cls, target: str,
                     rename_before: str,
                     rename_after: str,
                     return_data: bool = False,
                     must_affect_at_least_one: bool = True,
                     cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.renameField,
                   rename_before=rename_before,
                   rename_after=rename_after,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   cause=cause)

    @classmethod
    def count(cls, target: str,
              cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.count,
                   cause=cause)

    @classmethod
    def clear(cls, target: str,
              must_affect_at_least_one: bool = True,
              reset_serial: bool = False,
              cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.clear,
                   must_affect_at_least_one=must_affect_at_least_one,
                   reset_serial=reset_serial,
                   cause=cause)

    @classmethod
    def clear_add(cls, target: str,
                  add_data: List[CloneableFile],
                  return_data: bool = False,
                  must_affect_at_least_one: bool = True,
                  serial_key: Optional[str] = None,
                  reset_serial: bool = False,
                  cause: Optional[Cause] = None):
        return cls(target, EnumQueryType.clearAdd,
                   add_data=add_data,
                   return_data=return_data,
                   must_affect_at_least_one=must_affect_at_least_one,
                   serial_key=serial_key,
                   reset_serial=reset_serial,
                   cause=cause)

    # --------------------------
    # build() : Query インスタンスを返す
    # --------------------------
    def build(self) -> Query:
        m_data: Optional[List[Dict[str, Any]]] = None
        if self.add_data is not None:
            m_data = [i.to_dict() for i in self.add_data]

        return Query(
            target=self.target,
            type_=self.type,
            add_data=m_data,
            override_data=self.override_data,
            template=self.template,
            query_node=self.query_node,
            sort_obj=self.sort_obj,
            offset=self.offset,
            start_after=self.start_after,
            end_before=self.end_before,
            rename_before=self.rename_before,
            rename_after=self.rename_after,
            limit=self.limit,
            return_data=self.return_data,
            must_affect_at_least_one=self.must_affect_at_least_one,
            serial_key=self.serial_key,
            reset_serial=self.reset_serial,
            cause=self.cause
        )



from typing import Any, Dict, List, Optional

from file_state_manager.util_object_hash import UtilObjectHash

from delta_trace_db.db.util_copy import UtilCopy
from delta_trace_db.query.cause.enum_actor_type import EnumActorType
from delta_trace_db.query.cause.permission import Permission


# 深いコレクション比較（Dart の DeepCollectionEquality 相当）
def deep_collection_equals(a: Any, b: Any) -> bool:
    if isinstance(a, dict) and isinstance(b, dict):
        if a.keys() != b.keys():
            return False
        return all(deep_collection_equals(a[k], b[k]) for k in a)
    elif isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)):
        if len(a) != len(b):
            return False
        return all(deep_collection_equals(x, y) for x, y in zip(a, b))
    else:
        return a == b


class Actor:
    className = "Actor"
    version = "3"

    def __init__(
            self,
            actor_type: EnumActorType,
            actor_id: str,
            roles: List[str],
            permissions: List[str],
            collection_permissions: Optional[Dict[str, Permission]] = None,
            context: Optional[Dict[str, Any]] = None,
    ):
        self.actor_type = actor_type
        self.actor_id = actor_id
        self.roles = roles
        self.permissions = permissions
        self.collection_permissions = collection_permissions
        self.context = context

    @classmethod
    def from_dict(cls, src: Dict[str, Any]) -> "Actor":
        m_collection_permissions: Optional[Dict[str, Dict[str, Any]]] = None
        if src.get("collectionPermissions") is not None:
            m_collection_permissions = {
                k: v for k, v in src["collectionPermissions"].items()
            }

        collection_permissions: Optional[Dict[str, Permission]] = None
        if m_collection_permissions is not None:
            collection_permissions = {
                key: Permission.from_dict(value)
                for key, value in m_collection_permissions.items()
            }

        return cls(
            actor_type=EnumActorType[src["type"]],
            actor_id=src["id"],
            roles=list(src.get("roles", [])),
            permissions=list(src.get("permissions", [])),
            collection_permissions=collection_permissions,
            context=src.get("context"),
        )

    def clone(self) -> "Actor":
        return Actor.from_dict(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        m_collection_permissions: Optional[Dict[str, Dict[str, Any]]] = None
        if self.collection_permissions is not None:
            m_collection_permissions = {
                key: value.to_dict()
                for key, value in self.collection_permissions.items()
            }

        return {
            "className": self.className,
            "version": self.version,
            "type": self.actor_type.name,
            "id": self.actor_id,
            "roles": self.roles,
            "permissions": self.permissions,
            "collectionPermissions": m_collection_permissions,
            "context": UtilCopy.jsonable_deep_copy(self.context),
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Actor):
            return False
        return (
                self.actor_type == other.actor_type
                and self.actor_id == other.actor_id
                and self.roles == other.roles  # 順序あり比較
                and self.permissions == other.permissions  # 順序あり比較
                and deep_collection_equals(
            self.collection_permissions, other.collection_permissions
        )
                and deep_collection_equals(self.context, other.context)
        )

    def __hash__(self) -> int:
        return hash((
            self.actor_type,
            self.actor_id,
            UtilObjectHash.calc_list(self.roles),
            UtilObjectHash.calc_list(self.permissions),
            UtilObjectHash.calc_map(self.context),
        ))
